import os

from cndi.annotations.component import ComponentClass
import logging

from cndi.env import RCN_ACTIVE_PROFILE

logger = logging.getLogger("cndi.annotations")

beans = list()
autowires = list()
components = list()
beanStore = dict()
componentStore = dict()
profilesStores = dict()
conditionalRender = dict()
overrideStore = dict()

from functools import wraps
import importlib
import copy


def importModuleName(fullname):
    modules = fullname.split('.')
    module = importlib.import_module(modules[-1], package='.'.join(modules[:-1]))
    return module


def normaliseModuleAndClassName(name):
    nameList: list = name.split(".")
    if "__init__" in nameList:
        nameList.remove("__init__")
    return '.'.join(nameList)


def getBeanObject(objectType):
    bean = queryBeanStorage(objectType)
    objectInstance = bean['objectInstance']
    # if (objectInstance.__class__.__name__ == "function"):
    #     args[key] = objectInstance()
    # else:
    return copy.deepcopy(objectInstance) if bean['newInstance'] else objectInstance


def queryBeanStorage(fullname):
    objectType = normaliseModuleAndClassName(fullname)
    bean = beanStore[objectType]
    return bean


class AutowiredClass:
    def __init__(self, required, func, kwargs: dict()):
        self.fullname = '.'.join([func.__qualname__])
        self.className = normaliseModuleAndClassName('.'.join(func.__qualname__.split(".")[:-1]))
        self.func = func
        self.kwargs = kwargs
        self.required = required

    def dependencyInject(self):
        dependencies = self.calculateDependencies()
        dependencyNotFound = list()
        for dependency in dependencies:
            if dependency not in beanStore:
                dependencyNotFound.append(dependency)

        if len(dependencyNotFound) > 0:
            logger.warning(f"Skipping {self.fullname}")
            assert not self.required, "Could not initialize " + self.fullname + " with beans " + str(
                dependencyNotFound)

        kwargs = self.kwargs
        args = dict()
        for (key, value) in kwargs.items():
            fullName = normaliseModuleAndClassName('.'.join([value.__module__, value.__name__]))
            if fullName in beanStore:
                args[key] = getBeanObject(fullName)

        if self.className in beanStore:
            self.func(beanStore[self.className], **args)
        else:
            logger.info(f"{self.className} {self.fullname}")
            self.func(**args)

    def calculateDependencies(self):
        return list(
            map(lambda dependency: normaliseModuleAndClassName('.'.join([dependency.__module__, dependency.__name__])),
                self.kwargs.values()))


def OverrideBeanType(type: object):
    """
    OverrideBeanType is used to override the current annotated @Component class to be injected as some other object

    :param type: Class type to override while performing Dependency Injection
    :return: function wrapper
    """

    def inner_function(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        fullname = '.'.join([wrapper.__module__, wrapper.__name__])

        overrideStore[fullname] = {
            "func": wrapper,
            "overrideType": type
        }
        return wrapper

    return inner_function


def queryOverideBeanStore(fullname):
    if fullname in overrideStore:
        return overrideStore[fullname]
    else:
        return None


def Component(func: object):
    """
    When decorated with @Component, AppInitializer tries to automatically initialise the class this decorator is added to
    :param func:
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    moduleName = wrapper.__module__[:-9] if wrapper.__module__.endswith(".__init__") else wrapper.__module__
    componentFullName = '.'.join([moduleName, wrapper.__qualname__])
    logger.info(f"Registering Function name " + componentFullName)
    duplicateComponents = list(filter(lambda component: component.fullname == componentFullName, components))
    if duplicateComponents.__len__() > 0:
        logger.info(f"Duplicate Component found for: {duplicateComponents}")
    else:
        components.append(ComponentClass(**{
            'fullname': componentFullName,
            'func': wrapper,
            'annotations': wrapper.__init__.__annotations__ if "__annotations__" in dir(wrapper.__init__) else {}
        }))
    return wrapper


def validateBean(fullname):
    """
    Validates Beans before performing the dependency injection
    :param fullname: bean class full classpath name
    :return: Boolean type
    """
    profile = queryProfileData(fullname)
    condition = queryContitionalRenderingStore(fullname)
    if profile is None and condition is None:
        return True
    flag = True

    if profile is not None:
        profileNames = set(profile['profiles'])
        environmentProfiles = set(map(lambda x: x.strip(), os.environ[RCN_ACTIVE_PROFILE].split(",")))
        intersectionProfiles = profileNames.intersection(environmentProfiles)

        flag &= intersectionProfiles.__len__() >= 1

    if condition is not None:
        callback = condition['callback']
        callbackValue = callback(condition['func'])
        flag &= bool(callbackValue)

    if flag is False:
        logger.info("Validation Failed for Bean " + fullname)

    return flag


def Bean(newInstance=False):
    """

    :param newInstance:
    :return:
    """

    def inner_function(func):
        annotations = func.__annotations__
        returnType = annotations['return']
        del annotations['return']

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        fullname = '.'.join([returnType.__module__, returnType.__name__])
        annotations = dict(
            map(lambda key: (key, '.'.join([annotations[key].__module__, annotations[key].__qualname__])), annotations))

        if validateBean(fullname):
            beans.append({
                'name': fullname,
                'newInstance': newInstance,
                'object': wrapper,
                'fullname': wrapper.__qualname__,
                'kwargs': annotations,
                'index': len(beans)
            })

            return wrapper
        else:
            return None

    return inner_function


def ConditionalRendering(callback=lambda method: True):
    """

    :param callback:
    :return:
    """

    def inner_function(func: object):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        fullname = ".".join([wrapper.__module__, wrapper.__qualname__])
        conditionalRender[fullname] = {
            "func": wrapper,
            "callback": callback
        }

        return wrapper

    return inner_function


def queryContitionalRenderingStore(fullname):
    if fullname in conditionalRender:
        return conditionalRender[fullname]
    else:
        return None


def Profile(profiles=["default"]):
    """

    :param profiles:
    :return:
    """

    def inner_function(func: object):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        fullname = ".".join([wrapper.__module__, wrapper.__qualname__])
        profilesStores[fullname] = {
            "func": wrapper,
            "profiles": profiles
        }

        return wrapper

    return inner_function


def queryProfileData(fullname):
    if fullname in profilesStores:
        return profilesStores.get(fullname)
    else:
        return None


def Autowired(required=True):
    """
    
    :param required:
    :return:
    """

    def inner_function(func: object):
        annotations = func.__annotations__

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        autowires.append(AutowiredClass(required=required, **{
            'kwargs': annotations,
            'func': wrapper
        }))

        return wrapper

    return inner_function


def getBean(beans, name):
    return list(filter(lambda x: x['name'] == name, beans))[0]


def workOrder(beans):
    allBeanNames = list(map(lambda bean: bean['name'], beans))
    beanQueue = list(filter(lambda bean: len(bean['kwargs']) == 0, beans))
    beanIndexes = list(map(lambda bean: bean['index'], beanQueue))

    beanDependents = list(filter(lambda bean: bean['index'] not in beanIndexes, beans))
    beanQueueNames = list(map(lambda bean: bean['name'], beanQueue))

    for i in range(len(beanQueue)):
        beanQueue[i]['index'] = i

    for dependents in beanDependents:
        args = list(dependents['kwargs'].values())
        flag = True
        for argClassName in args:
            if (argClassName not in beanQueueNames and argClassName in allBeanNames) or argClassName in beanQueueNames:
                flag = flag and True
                dependents['index'] = getBean(beans, argClassName)['index'] + max(beanIndexes)
            else:
                flag = False

        if flag:
            beanQueue.append(dependents)
            beanQueueNames.append(dependents['name'])

    assert len(beanQueue) == len(beans), "Somebeans were not initialized properly"
    return list(sorted(beanQueue, key=lambda x: x['index']))



from paho.mqtt.client import Client

from cndi.binders.message.utils import MessageChannel


class MqttProducerBinding(MessageChannel):
    def __init__(self, mqttClient: Client):
        self.mqttClient = mqttClient
        self.topic = None

    def send(self, message) -> None:
        self.mqttClient.publish(self.topic, message)

    def close(self):
        self.mqttClient.loop_stop(force=True)


from pathlib import Path

from yaml import SafeLoader, load_all
import os, re
import logging

from cndi import BASE_NAME

logger = logging.getLogger(__name__)


RCN_ENVS_CONFIG = f'{BASE_NAME}_ENVS_CONFIG'
RCN_ACTIVE_PROFILE = f"{RCN_ENVS_CONFIG}.active.profile" # RCN_ENVS_CONFIG.active.profile


if f"{BASE_NAME}_HOME" not in os.environ:
    os.environ[f"{BASE_NAME}_HOME"] = f"{Path.home()}/.rcn/"

configDir = os.environ[f"{BASE_NAME}_HOME"]

if RCN_ACTIVE_PROFILE not in os.environ:
    os.environ[RCN_ACTIVE_PROFILE] = "default"

VARS = dict(map(lambda key: (key,os.environ[key]), filter(lambda key: key.startswith(RCN_ENVS_CONFIG), os.environ)))

def loadEnvsFromRcnHome():
    profile = getConfiguredProfile()
    profileHome = os.path.join(configDir, profile)
    profileEnvFile = os.path.join(profileHome, "env.yml")

    if os.path.exists(profileEnvFile):
        logger.info(f"Profile Environment file found for profile f{profile}")
        loadEnvFromFile(profileEnvFile)

def addToOsEnviron(key: str, value):
    if not key.startswith("."):
        key = '.' + key
    if (RCN_ENVS_CONFIG+key) not in VARS:
        VARS[(RCN_ENVS_CONFIG+key)] = str(value)
    else:
        logger.warning(f"An env variable already exists with key={(RCN_ENVS_CONFIG+key)}")

def walkListKey(parent, parent_label=''):
    responseList = list()
    for i,value in enumerate(parent):
        if isinstance(value, dict):
            responseList.extend(walkDictKey(value, parent_label + '.#' + str(i)))
        elif isinstance(value, list):
            responseList.extend(walkListKey(value, parent_label + '.#' + str(i)))
        else:
            responseList.append([parent_label + '.#'+ str(i), value])

    return responseList

def walkDictKey(parent, parent_label=''):
    responseList = list()
    for key, value in parent.items():
        if isinstance(value, dict):
            responseList.extend(walkDictKey(value, parent_label + '.' + key))
        elif isinstance(value, list):
            responseList.extend(walkListKey(value, parent_label + '.' + key))
        else:
            responseList.append([parent_label + '.'+ key, value])

    return responseList

def loadEnvFromFiles(*files):
    for file in files:
        if not os.path.exists(file):
            logger.info(f"Env file does not exist: {file}")
            continue

        loadEnvFromFile(file)

def getConfiguredProfile():
    if f"{RCN_ENVS_CONFIG}.active.profile".lower() in VARS:
        return VARS[f"{RCN_ENVS_CONFIG}.active.profile".lower()]
    elif "rcn.active.profile" in os.environ:
        return os.environ["rcn.active.profile"]
    else:
        return "default"

def loadEnvFromFile(property_file):
    if(not os.path.exists(property_file)):
        raise FileNotFoundError(f"Environment file does not exists at {property_file}")

    with open(property_file, "r") as stream:
        data = list(load_all(stream, SafeLoader))
        if len(data) == 1:
            data = data[0]
        else:
            dataDict = dict(map(lambda x: (x['rcn']['profile'], x), data))
            profile = getConfiguredProfile()
            data = dataDict[profile]
            data = normalize(data)
        envData = walkDictKey(data)
        for key, value in envData:
            addToOsEnviron(key, value)

def getContextEnvironments():
    return dict(
        map(
            lambda items: [items[0][RCN_ENVS_CONFIG.__len__()+1:].lower(), items[1]],
            filter(lambda items: items[0].startswith(RCN_ENVS_CONFIG), VARS.items())
        )
    )

def getListTypeContextEnvironments():
    rcn_envs = getContextEnvironments()
    dataDict = dict(filter(lambda key: key[0].__contains__(".#"), rcn_envs.items()))
    return dataDict

def getContextEnvironment(key: str, defaultValue = None, castFunc = None, required=False):
    envDict = getContextEnvironments()
    key = key.lower()
    if key in envDict:
        if castFunc is not None:
            return castFunc(envDict[key])
        return envDict[key]
    if required:
        raise KeyError(f"Environment Variable with Key: {key} not found")
    return defaultValue

def constructDictWithValues(value, keys=[]):
    if len(keys) == 1:
        return {
            keys[0]: value
        }
    else:
        return {
            keys[0]: constructDictWithValues(value, keys[1:])
        }

def constructDict(value, generatedObject, key=''):
    tempDict = generatedObject
    keys = key.split('.')
    for i, key in enumerate(keys):
        if key == "":
            continue

        if key in tempDict:
            tempDict = tempDict[key]
        else:
            tempDict.update(constructDictWithValues(value, keys[i:]))
            break

def normalize(dictObject: dict, key = ''):
    envData = walkDictKey(dictObject)
    envData = sorted(envData, key = lambda data: data[0])

    generatedDict = dict()
    for key, value in envData:
        searchResult = re.findall("\${[a-z0-9A-Z\\_]+}", str(value))
        for result in searchResult:
            groupValue = re.match("\${(?P<envName>[a-z0-9A-Z\\_]+)}", result).group('envName')
            if groupValue.startswith(RCN_ENVS_CONFIG) and groupValue in VARS:
                envValue = VARS[groupValue]
            elif groupValue in os.environ:
                envValue = os.environ[groupValue]
            else:
                raise ValueError(f"Environment Variable not found {groupValue}")
            value = value.replace(result, envValue)

        constructDict(value, generatedDict, key)

    return generatedDict



from functools import wraps

from cndi.annotations import getBeanObject, Component, ConditionalRendering
from cndi.binders.consts import RCN_MESSAGE_BINDERS
from cndi.binders.message.utils import extractChannelNameFromPropertyKey, SubscriberChannel
from cndi.env import getContextEnvironment, getContextEnvironments
import logging

logger = logging.getLogger(__name__)


CHANNELS_TO_TOPIC_MAP = dict()
CHANNELS_TO_FUNC_MAP = dict()


class Message:
    def __init__(self, message):
        self.message = message
        self.key = None

    def setMessage(self, message):
        self.message = message
        return self

    def setKey(self, key):
        self.key = key
        return self

def Input(channelName):
    def inner_function(func):
        CHANNELS_TO_FUNC_MAP[channelName] = dict(func=func, annotations=func.__annotations__, is_sink=False)
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return inner_function

def Output(channelName):
    def inner_function(func):
        CHANNELS_TO_FUNC_MAP[channelName] = dict(func=func, annotations=func.__annotations__, is_sink=True)
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return inner_function

def _conditionalRenderDefaultMessageBinder(wrapper):
    defaultMessageBinder = getContextEnvironment("rcn.binders.message.enable",  defaultValue=False)
    return defaultMessageBinder

@Component
@ConditionalRendering(callback=_conditionalRenderDefaultMessageBinder)
class DefaultMessageBinder:
    def   __init__(self):
        self.logger = logging.getLogger('.'.join([self.__class__.__module__, self.__class__.__name__]))

        self.binders = dict()
        self.topicConsumers = dict()
        self.callbacks = list()
        self.defaultMessageBinder = getContextEnvironment("rcn.binders.message.default")
        self.initializeBinders()

    def start(self):
        for callback in self.callbacks:
            callback()

    def performInjection(self):
        for channelName, methodWrapper in CHANNELS_TO_FUNC_MAP.items():
            if channelName not in self.binders:
                self.logger.error(f"No binding found for channel - {channelName}")
                continue

            if methodWrapper['is_sink']:
                binder = self.binders[channelName]
                method = methodWrapper['func']
                methodsClassFullname = f"{method.__module__}.{method.__qualname__.split('.')[0]}"
                classBean = getBeanObject(methodsClassFullname)
                method(classBean, binder)

    def initializeBinders(self):
        contextEnvs = getContextEnvironments()

        if self.defaultMessageBinder.strip().lower() == "rabbitmq":
            from cndi.binders.message.rabbitmq import RabbitMQProducerBinding, RabbitMQBinder

            rabbitMqBinder = RabbitMQBinder()

            self.binders.update(rabbitMqBinder.bindProducers())
            self.binders.update(rabbitMqBinder.bindSubscribers(CHANNELS_TO_FUNC_MAP=CHANNELS_TO_FUNC_MAP))
            self.callbacks.append(rabbitMqBinder.channelThread.start)

        elif self.defaultMessageBinder.strip().lower() == "mqtt":
            from cndi.binders.message.mqtt import MqttProducerBinding
            from paho.mqtt.client import Client, MQTTMessage

            brokerUrl = getContextEnvironment(f"{RCN_MESSAGE_BINDERS}.mqtt.brokerUrl", defaultValue=None)
            brokerPort = getContextEnvironment(f"{RCN_MESSAGE_BINDERS}.mqtt.brokerPort", defaultValue=None)

            brokerUrl = getContextEnvironment(f"{RCN_MESSAGE_BINDERS}.brokerUrl", required=True, defaultValue=brokerUrl)
            brokerPort = getContextEnvironment(f"{RCN_MESSAGE_BINDERS}.brokerPort", required=True, castFunc=int, defaultValue=brokerPort)

            mqttClient = Client()

            mqttProducerChannelBindings = filter(lambda key: key.startswith('rcn.binders.message.mqtt.producer'), contextEnvs)
            for propertyKey in mqttProducerChannelBindings:
                channelName = extractChannelNameFromPropertyKey(propertyKey)
                producerBinding = MqttProducerBinding(mqttClient)
                topicName = getContextEnvironment(propertyKey, required=True)

                producerBinding.setTopic(topicName)
                CHANNELS_TO_TOPIC_MAP[channelName] = topicName
                self.binders[channelName] = producerBinding

            mqttConsumerChannelBindings = filter(lambda key: key.startswith('rcn.binders.message.mqtt.consumer') and key.endswith("destination"), contextEnvs)
            subscriptionTopics = list()

            for propertyKey in mqttConsumerChannelBindings:
                channelName = extractChannelNameFromPropertyKey(propertyKey)
                if channelName not in CHANNELS_TO_FUNC_MAP:
                    self.logger.error(f"Channel not found: {channelName}")
                    continue
                consumerBinding = SubscriberChannel()
                callbackDetails = CHANNELS_TO_FUNC_MAP[channelName]
                callback = callbackDetails['func']
                consumerBinding.setOnConsumeCallback(callback)
                topicName = getContextEnvironment(propertyKey, required=True)
                consumerBinding.setTopic(topicName)
                subscriptionTopics.append(topicName)
                if topicName in self.topicConsumers:
                    raise KeyError(f"Duplicate topic found {topicName} with {self.topicConsumers[topicName]}")

                self.topicConsumers[topicName] = consumerBinding
                self.binders[channelName] = consumerBinding

            def on_connect(client: Client, userdata, flags, rc):
                self.logger.info("Connected with result code " + str(rc))
                for topic in subscriptionTopics:
                    client.subscribe(topic)

            def on_message(client, userdata, msg: MQTTMessage):
                self.topicConsumers[msg.topic](msg)

            mqttClient.on_connect = on_connect
            mqttClient.on_message = on_message

            mqttClient.connect(brokerUrl, brokerPort)
            self.callbacks.append(mqttClient.loop_start)